//
//  Smart Pointers:
//  - std::shared_ptr
//  - std::weak_ptr
//  jsonserial - Eric Lecolinet
// See: https://www.telecom-paris.fr/~elc/jsonserial/
//

#include "main.hpp"
#include "../jsonserial.hpp"
#include "../jsonserial_list.hpp"
#include "../jsonserial_map.hpp"
using namespace jsonserial;

namespace smart_pointers {

using ContactPtr = std::shared_ptr<class Contact>;

class Contact {
public:
  enum Gender {Unknown, Male, Female, Other};
  struct Phone {std::string type, number;};

  friend JSonClasses* registerClasses();  // needed if variables are not public!

  Contact() = default;                        // needed to instantiate pointees!

  Contact(std::string const& firstname, std::string const& lastname,
          unsigned int age, Gender gender)
  : firstname(firstname), lastname(lastname), age(age), gender(gender) {}

  void addPhone(std::string const& type, std::string const& number) {
    phones.push_back(new Phone{type, number});
  }

  static void linkPartners(ContactPtr p1, ContactPtr p2)  {
    if (p1) p1->partner = p2;
    if (p2) p2->partner = p1;
  }
  
  static void linkChildWithParents(ContactPtr child, ContactPtr p1, ContactPtr p2) {
    if (child) {
      child->parent1 = p1;
      child->parent2 = p2;
      if (p1) p1->children[child->firstname] = child;
      if (p2) p2->children[child->firstname] = child;
    }
  }

  void print(std::ostream& out) const {
    out
    << "First:\t" << firstname << "\n"
    << "Last:\t"  << lastname  << "\n"
    << "Gender:\t"<< gender << "\n"
    << "Age:\t" << age << "\n"
    << "Phones:\t";
    for (auto& it : phones) {
      out << "[" << it->type << ": " << it->number << "] ";
    }
    out << "\n";
    if (auto p = partner.lock() ) out << "Partner:\t"<< p->firstname << "\n";
    if (parent1) out << "Parent1:\t"<< parent1->firstname << "\n";
    if (parent2) out << "Parent2:\t"<< parent2->firstname << "\n";
    if (!children.empty()) {
      out << "Children:\t";
      for (auto& it : children) {
        if (auto p = it.second.lock()) out << p->firstname << " ";
      }
      out << "\n";
    }
  }

private:
  std::string firstname, lastname;
  Gender gender;
  unsigned int age{};
  std::list<Phone*> phones;
  std::weak_ptr<Contact> partner;
  std::shared_ptr<Contact> parent1, parent2;
  std::map<std::string, std::weak_ptr<Contact>> children;
};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

JSonClasses* registerClasses() {
  JSonClasses* classes = new JSonClasses();

  classes->addClass<Contact>("Contact")
  .addMember("firstname", &Contact::firstname)
  .addMember("lastname", &Contact::lastname)
  .addMember("gender", &Contact::gender)
  .addMember("age", &Contact::age)
  .addMember("phones", &Contact::phones)
  .addMember("partner", &Contact::partner)
  .addMember("parent1", &Contact::parent1)
  .addMember("parent2", &Contact::parent2)
  .addMember("children", &Contact::children);

  classes->addClass<Contact::Phone>("Phone")
  .addMember("type", &Contact::Phone::type)
  .addMember("number", &Contact::Phone::number);

  return classes;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

using Contacts = std::list<ContactPtr>;

void printList(std::string const& title, Contacts const& list) {
  std::cout << "--- " << title << " --------\n\n";
  for (auto& it : list) {it->print(std::cout); std::cout << std::endl;}
}

Contacts createList() {
  ContactPtr bob = std::make_shared<Contact>("Bob", "Dûpontèle", 31, Contact::Male);
  bob->addPhone("home", "123 456-7890");
  bob->addPhone("mobile","703 221-2121");

  ContactPtr alice = std::make_shared<Contact>("Alice", "Dûpontèle", 33, Contact::Female);
  alice->addPhone("home", "123 456-7890");
  alice->addPhone("office", "211 1234-1234");

  ContactPtr karim = std::make_shared<Contact>("Karim", "Dûpontèle", 9, Contact::Male);
  karim->addPhone("mobile","122 122-1222");

  ContactPtr susan = std::make_shared<Contact>("Susan", "Dûpontèle", 11, Contact::Female);
  susan->addPhone("mobile","133 133-1333");

  Contact::linkPartners(bob, alice);
  Contact::linkChildWithParents(karim, bob, alice);
  Contact::linkChildWithParents(susan, bob, alice);

  return Contacts{bob, alice, karim, susan};
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

static JSonSerial js(registerClasses());


void test_file(std::string const& filename) {
  Contacts list = createList();
  printList("Original List", list);

  // CAUTION: last argument must be true because the objects' graph is cyclic!
  if (!js.write(list, filename, true)) return;

  Contacts list2;
  if (!js.read(list2, filename)) return;
  printList("Copy after Reading", list2);
}


void test_stream() {
  Contacts list = createList();
  printList("Original List", list);

  std::stringstream buffer;

  // CAUTION: last argument must be true because the objects' graph is cyclic!
  if (!js.write(list, buffer, true)) return;
  std::cout << buffer.str() << std::endl;

  Contacts list2;
  if (!js.read(list2, buffer)) return;
  printList("Copy after Reading", list2);
}

}
